import types
from functools import partial
from .router import Router


def get_module_fns(module):
    "get defined functions of a module"
    attrs = [getattr(module, a) for a in dir(module) if not a.startswith('_')]
    return [attr for attr in attrs if isinstance(attr, types.FunctionType)
            and attr.__module__ == module.__name__]


def get_methods(cls):
    "get public methods of a class"
    attrs = [getattr(cls, a) for a in dir(cls) if not a.startswith('_')]
    return [attr for attr in attrs if isinstance(attr, types.FunctionType)]


class RestfulRouter(Router):

    restful_routes = [
        ('GET',    '%(path)s',               'query'),
        ('POST',   '%(path)s',               'create'),
        ('GET',    '%(path)s/<%(id)s>',      'show'),
        ('PUT',    '%(path)s/<%(id)s>',      'replace'),
        ('PATCH',  '%(path)s/<%(id)s>',      'modify'),
        ('DELETE', '%(path)s/<%(id)s>',      'destroy'),
        ('GET',    '%(path)s/new',           'new'),
        ('GET',    '%(path)s/<%(id)s>/edit', 'edit'),
    ]

    restful_methods = [method for _, _, method in restful_routes]

    def resource(self, url_path=None, module=None):
        """register a restful resource

        Example:

            @resource('/article')
            class Article:
                def show(article_id):
                    pass

                @method('put')
                def upvote(article_id):
                    pass

        """
        if url_path is None:
            url_path = '/' + module.__name__.replace('.', '/')
        if module is None:
            return partial(self.register_resource, url_path)
        else:
            return self.register_resource(url_path, module)

    def resources(self, *resources, prefix=''):
        """register a list of restful resources

        Example:

            resources(articles, users, '/api/v1')

        """
        for resource in resources:
            url_path = prefix + '/' + resource.__name__.split('.').pop()
            self.register_resource(url_path, resource)

    def register_resource(self, url_path, module):
        url_id = '%s_id' % url_path.split('/').pop()
        vals = {'path': url_path, 'id': url_id}
        rules = [(method, pattern % vals, getattr(module, handler))
                 for method, pattern, handler in self.restful_routes
                 if hasattr(module, handler)]

        if isinstance(type(module), types.ModuleType):
            fns = get_module_fns(module)
        else:
            fns = get_methods(module)

        custom_rules = [(getattr(fn, '__httpmethod__', 'GET'),
                        '%s/<%s>/%s' % (url_path, url_id, fn.__name__),
                        fn) for fn in fns
                        if fn.__name__ not in self.restful_methods]
        self.extend(rules)
        self.extend(custom_rules)
        return module



import re
from collections import defaultdict


class Router:

    param_macher = re.compile(r"\(\?P<([^>]+)>[^)]*\)")

    def __init__(self):
        self.__idx__ = defaultdict(list)
        self.__reversedidx__ = {}

    def append(self, method, pattern, handler):
        if type(pattern) is str:
            pattern = self.parse_pattern(pattern)
        self.__idx__[self.index(pattern)].append((method, pattern, handler))
        name = handler if type(handler) is str else handler.__qualname__
        self.__reversedidx__[name] = pattern
        return handler

    def extend(self, rules):
        for rule in rules:
            self.append(*rule)

    def index(self, pattern):
        pattern = pattern.pattern.lstrip('^').rstrip('$')
        pos = pattern.find('(')
        if pos != -1:
            pattern = pattern[:pos]
        segments = pattern.split('/')
        segments.pop()
        return '/'.join(segments)

    def parse_pattern(self, pattern):
        pattern = re.compile('^%s$' %
                             re.sub(r'<([^>]+)>', r'(?P<\1>[^/]+)', pattern))
        return pattern

    def match(self, method, path):
        """find handler from registered rules

        Example:

            handler, params = match('GET', '/path')

        """
        segments = path.split('/')
        while len(segments):
            index = '/'.join(segments)
            if index in self.__idx__:
                return self.match_rule(method, path, self.__idx__[index])
            segments.pop()
        return None, None

    def match_rule(self, method, path, rules):
        for _method, pattern, handler in rules:
            if _method != method:
                continue
            result = pattern.match(path)
            if result:
                return handler, result.groupdict()
        return None, None

    def path_for(self, handler, **kwargs):
        """construct path for a given handler

        Example:

            path = path_for(show_user, user_id=109)
        """
        if type(handler) is not str:
            handler = handler.__qualname__
        if handler not in self.__reversedidx__:
            return None
        pattern = self.__reversedidx__[handler].pattern.lstrip('^').rstrip('$')
        return self.param_macher.sub(lambda m: str(kwargs[m.group(1)]),
                                     pattern)

    def __repr__(self):
        return "\n".join(["%s %s -> %s" %
                          (method, pattern.pattern, handler.__qualname__)
                          for rules in self.__idx__.values()
                          for (method, pattern, handler) in rules])



from setuptools import setup

setup(
    name='biro',
    version='0.0.1',
    url='https://github.com/zweifisch/biro',
    license='MIT',
    description='bidirectional URI routing',
    keywords='bidirectional routing',
    long_description=open('README.md').read(),
    author='Feng Zhou',
    author_email='zf.pascal@gmail.com',
    packages=['biro'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.3',
        'Topic :: Internet :: WWW/HTTP :: WSGI'
    ]
)



"""
bidirectional URI routing
"""

from functools import partial
from .restful import RestfulRouter

routes = RestfulRouter()

path_for = routes.path_for
match = routes.match
resource = routes.resource
resources = routes.resources


def method(httpmethod):
    """decorator to overwrite default method(GET) for custom actions,
    intended to be used within restful resource
    """
    def add_method(fn):
        fn.__httpmethod__ = httpmethod
        return fn
    return add_method


def route(method, pattern, handler=None):
    """register a routing rule

    Example:

         route('GET', '/path/<param>', handler)

    """
    if handler is None:
        return partial(route, method, pattern)
    return routes.append(method, pattern, handler)

for _ in ["get", "post", "head", "delete", "put", "patch"]:
    locals()[_] = partial(route, _.upper())
    locals()[_].__doc__ = """register a %(method)s handler

    Example:

        %(method)s('/path/<param', handler)

        @%(method)s('/path/<param')
        def handler(param):
            pass
    """ % dict(method=_)
